(function () {
'use strict';

var indices = ['First', 'Second', 'Third'];

function validateArguments(args, func, options) {
  Object.keys(options).map(function (name, index) {
    var ref = options[name];
    var expected = ref[0];
    var required = ref[1];
    var value = args[index];
    var type = Array.isArray(value) ? 'array' : typeof value;
    var matches = type === expected;

    return invariant(
      required ? matches : value === undefined || matches,
      ((indices[index]) + " argument of " + func + " should be the " + name + " of type '" + expected + "', got '" + (typeof value) + "'.")
    )
  });
}

function invariant(condition, message) {
  if (!condition) {
    throw new Error(message)
  }
}

/**
 * @typedef   {Object}    Container
 * @property  {Function}  add
 * @property  {Function}  share
 * @property  {Function}  get
 */

/**
 * @return {Container}
 */
function createContainer() { // eslint-disable-line max-statements
  var dependencies = [];
  var subscribers = [];

  /**
   * Registers a depedency at the container.
   *
   * @param {string}    name      Dependency name
   * @param {Function}  func      Constructor or factory function
   * @param {Array}     [args=[]] Dependency arguments to inject
   */
  function add(name, func, args) {
    if ( args === void 0 ) args = [];

    validateRegistrationArguments(name, func, args, 'add');
    var dependency = { name: name, func: func, args: args };
    dependencies = dependencies.concat( [dependency]);
    subscribers.forEach(function (subscriber) { return subscriber(); });
  }

  /**
   * Registers a singleton depedency at the container.
   *
   * @param {string}    name      Dependency name
   * @param {Function}  func      Constructor or factory function
   * @param {Array}     [args=[]] Dependency arguments to inject
   */
  function share(name, func, args) {
    if ( args === void 0 ) args = [];

    validateRegistrationArguments(name, func, args, 'share');
    var dependency = { name: name, func: func, args: args, shared: true };
    dependencies = dependencies.concat( [dependency]);
    subscribers.forEach(function (subscriber) { return subscriber(); });
  }

  /**
   * Get an instantiated, injected dependency from the container.
   *
   * @param {string}  name  Dependency name
   * @return {*}  Resolved dependency
   */
  function get(name) {
    validateArguments([name], 'get', { name: ['string', true] });
    var dependency = findDependency(name);
    invariant(dependency, ("Dependency '" + name + "' does not exist."));
    return instantiate(dependency)
  }

  function lazy(name, cb) {
    subscribers = subscribers.concat( [function () {
      var dependency = findDependency(name);

      if (dependency) {
        return cb(instantiate(dependency))
      }
    }]);
  }

  function instantiate(ref) {
    var name = ref.name;
    var func = ref.func;
    var instance = ref.instance;
    var args = ref.args;
    var shared = ref.shared;

    if (shared) {
      return instance || createSingle(name, func, instance, args)
    }

    return create(func, args)
  }

  function create(func, args) {
    var resolvedArgs = args.map(get);
    return func.hasOwnProperty('prototype') ?
      new (Function.prototype.bind.apply( func, [ null ].concat( resolvedArgs) )) : // eslint-disable-line new-cap
      func.apply(void 0, resolvedArgs)
  }

  function createSingle(name, func, instance, args) {
    var newInstance = create(func, args);
    dependencies = setInstance(name, newInstance);
    return newInstance
  }

  function findDependency(name) {
    return dependencies.find(function (dep) { return dep.name === name; })
  }

  function setInstance(name, instance) {
    return dependencies.map(function (dep) { return name === dep.name ?
        Object.assign({}, dep, {instance: instance}) :
        dep; }
    )
  }

  function validateRegistrationArguments(name, func, args, type) {
    validateArguments([name, func, args], type, {
      name: ['string', true],
      constructor: ['function', true],
      arguments: ['array']
    });
    invariant(!findDependency(name), ("Dependency '" + name + "' already exists."));
  }

  return { add: add, share: share, get: get, lazy: lazy }
}

/* global window */

window.createContainer = createContainer;

}());
//# sourceMappingURL=containor.js.map
